import os
import argparse
from pathlib import Path
import time
from tqdm import tqdm
from exordium.video.facedetector import RetinaFaceDetector
from exordium.video.tracker import IouTracker
from exordium.video.fabnet import FabNetWrapper
from exordium.video.opengraphau import OpenGraphAuWrapper
from exordium.video.openface import extract_openface_singularity

DB = Path("data/db_processed/mosei")
DB_RAW_VIDEOS = DB / "Raw"
DB_VIDEOS = DB / "Converted"


def reencode_video(raw_video_path):
    converted_video_path = DB_VIDEOS / raw_video_path.parent.name / raw_video_path.name
    converted_video_path.parent.mkdir(parents=True, exist_ok=True)
    if converted_video_path.exists(): return
    os.system(f'ffmpeg -y -i {raw_video_path} -r 30 {converted_video_path}')


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Script to preprocess MOSI visual features.")
    parser.add_argument('--gpu_id', type=int, default=0, help='ID of the GPU to use')
    parser.add_argument('--start',  type=int, default=0, help='participant id slice start')
    parser.add_argument('--end',    type=int, default=23000, help='participant id slice end')
    args = parser.parse_args()

    '''
    for raw_video_path in tqdm(sorted(list(DB_RAW_VIDEOS.glob("**/*.mp4")))):
        reencode_video(raw_video_path)
    '''

    print(f"Using GPU ID: {args.gpu_id}")
    face_detector = RetinaFaceDetector(gpu_id=args.gpu_id, batch_size=10)
    fabnet_extractor = FabNetWrapper(gpu_id=args.gpu_id)
    au_extractor = OpenGraphAuWrapper(gpu_id=args.gpu_id)

    video_paths = sorted(list(DB_VIDEOS.glob("**/*.mp4")))[args.start:args.end]
    print('Number of videos:', len(video_paths))

    for v, video_path in tqdm(enumerate(video_paths), total=len(video_paths), desc='Videos'):
        start_time = time.time()
        video_name = video_path.parent.name
        video_id = video_path.stem

        try:
            print("video:", video_path)
            videodetections = face_detector.detect_video(video_path, output_path=DB / 'tracker' / video_name / f'{video_id}.vdet')
            track = IouTracker(max_lost=30).label(videodetections).merge().select_topk_long_tracks(top_k=2).get_center_track()
            print("detected track length:", len(track))

            ids, embeddings = fabnet_extractor.track_to_feature(track, batch_size=30, output_path=DB / 'fabnet' / video_name / f'{video_id}.pkl')
            print('fabnet embeddings:', embeddings.shape)
        
            ids, au = au_extractor.track_to_feature(track, batch_size=30, output_path=DB / 'opengraphau' / video_name / f'{video_id}.pkl')
            print('au:', au.shape)

            if not (DB / 'openface' / video_name / video_id / f'{video_id}.csv').exists():
                extract_openface_singularity(
                    video_path,
                    DB / 'openface' / video_name / video_id,
                    str(Path('tools/openface_runner/openface_latest.sif').resolve()), 
                    'tools/openface_runner/tools',
                    singularity_args='--bind /nas:/nas')
        
        except Exception as e:
            with open(DB / "mosei_skip_video.txt", "a") as f:
                f.write(f'{video_name},{video_id},{e}\n')

        end_time = time.time()
        print(f"Elapsed time: {end_time - start_time:03f} seconds")


from pathlib import Path
import pickle
import numpy as np
import matplotlib.pyplot as plt
from performance_metrics import speedup_ratio, percentage_reduction

custom_blue = (0.855, 0.91, 0.988)
custom_green = (0.835, 0.91, 0.831)
line_width = 2
border_blue = (0.424, 0.557, 0.749)
border_green = (0.51, 0.702, 0.40)


def plot_inference_complexity(mult_results, linmult_results):
    # Example data (replace with your actual data)
    batch_sizes_mult = [1, 2, 3, 4, 5, 6, 7, 8]
    batch_sizes_linmult = [1, 2, 3, 4, 5, 6, 7, 8, 12, 16, 20, 24, 28, 32, 36, 40, 44, 48]

    # Inference time (ms) - replace with actual mean and std lists
    mult_inference_time_mean = [elem['inference_time_mean'] for elem in mult_results]
    mult_inference_time_std = [elem['inference_time_std'] for elem in mult_results]
    linmult_inference_time_mean = [elem['inference_time_mean'] for elem in linmult_results]
    linmult_inference_time_std = [elem['inference_time_std'] for elem in linmult_results]

    # FLOPs - replace with actual FLOPs values
    mult_flops = [elem['flops'] for elem in mult_results]
    linmult_flops = [elem['flops'] for elem in linmult_results]

    # Memory usage (MB) - replace with actual memory usage values
    mult_memory_allocated_usage = [elem['peak_memory_allocated'] for elem in mult_results]
    linmult_memory_allocated_usage = [elem['peak_memory_allocated'] for elem in linmult_results]
    mult_memory_reserved_usage = [elem['peak_memory_reserved'] for elem in mult_results]
    linmult_memory_reserved_usage = [elem['peak_memory_reserved'] for elem in linmult_results]

    # Speedup ratio and memory reduction - replace with actual calculated values
    speedup = [speedup_ratio(mult_elem['inference_time_mean'], linmult_elem['inference_time_mean']) for mult_elem, linmult_elem in zip(mult_results, linmult_results)]
    memory_reduction_allocated = [percentage_reduction(mult_elem['peak_memory_allocated'], linmult_elem['peak_memory_allocated']) for mult_elem, linmult_elem in zip(mult_results, linmult_results)]
    memory_reduction_reserved = [percentage_reduction(mult_elem['peak_memory_reserved'], linmult_elem['peak_memory_reserved']) for mult_elem, linmult_elem in zip(mult_results, linmult_results)]

    # Creating the figure with 4 subplots
    fig, axs = plt.subplots(2, 2, figsize=(14, 10))  # 2x2 grid of subplots
    fig.tight_layout(pad=3.0)  # Adjust padding between subplots

    # Plot 1: Inference time comparison
    axs[0, 0].errorbar(batch_sizes_mult, mult_inference_time_mean, yerr=mult_inference_time_std,
                    fmt='-o', color='blue', label='MulT', capsize=5)
    axs[0, 0].errorbar(batch_sizes_linmult, linmult_inference_time_mean, yerr=linmult_inference_time_std,
                    fmt='-o', color='green', label='LinMulT', capsize=5)
    axs[0, 0].set_title('Inference Time Comparison')
    axs[0, 0].set_xlabel('Batch Size')
    axs[0, 0].set_ylabel('Time (ms)')
    axs[0, 0].set_xticks(batch_sizes_linmult)
    axs[0, 0].set_xticklabels(batch_sizes_linmult)
    axs[0, 0].legend()
    axs[0, 0].grid()

    # Plot 2: FLOPs comparison
    axs[0, 1].plot(batch_sizes_mult, mult_flops, '-o', color='blue', label='MulT')
    axs[0, 1].plot(batch_sizes_linmult, linmult_flops, '-o', color='green', label='LinMulT')
    axs[0, 1].set_title('FLOPs Comparison')
    axs[0, 1].set_xlabel('Batch Size')
    axs[0, 1].set_ylabel('FLOPs')
    axs[0, 1].set_xticks(batch_sizes_linmult)
    axs[0, 1].set_xticklabels(batch_sizes_linmult)
    axs[0, 1].legend()
    axs[0, 1].grid()

    # Plot 3: Memory usage comparison
    axs[1, 0].plot(batch_sizes_mult, mult_memory_allocated_usage, '-o', color='blue', label='MulT')
    axs[1, 0].plot(batch_sizes_linmult, linmult_memory_allocated_usage, '-o', color='green', label='LinMulT')
    axs[1, 0].plot(batch_sizes_mult, mult_memory_reserved_usage, ':o', color='blue', label='MulT')
    axs[1, 0].plot(batch_sizes_linmult, linmult_memory_reserved_usage, ':o', color='green', label='LinMulT')
    axs[1, 0].set_title('Memory Usage Comparison')
    axs[1, 0].set_xlabel('Batch Size')
    axs[1, 0].set_ylabel('Memory Usage (MB)')
    axs[1, 0].set_xticks(batch_sizes_linmult)
    axs[1, 0].set_xticklabels(batch_sizes_linmult)
    axs[1, 0].legend()
    axs[1, 0].grid()

    # Plot 4: Speedup ratio and memory reduction comparison
    ax4 = axs[1, 1]
    ax4.plot(batch_sizes_mult, speedup, '-o', color='blue', label='Speedup Ratio')
    ax4.set_xlabel('Batch Size')
    ax4.set_ylabel('Speedup Ratio', color='blue')
    ax4.tick_params(axis='y', labelcolor='blue')
    ax4.set_xticks(batch_sizes_linmult)
    ax4.set_xticklabels(batch_sizes_linmult)
    ax4.grid()

    # Create a twin axis for memory reduction
    ax4b = ax4.twinx()
    ax4b.plot(batch_sizes_mult, memory_reduction_allocated, '-s', color='green', label='Memory allocated Reduction (%)')
    ax4b.plot(batch_sizes_mult, memory_reduction_reserved, ':s', color='green', label='Memory reserved Reduction (%)')
    ax4b.set_ylabel('Memory Reduction (%)', color='green')
    ax4b.set_ylim([0, 100])
    ax4b.set_yticks(list(range(0, 101, 10)))
    ax4b.tick_params(axis='y', labelcolor='green')

    # Title and legends for the comparison plot
    ax4.set_title('Speedup Ratio and Memory Reduction Comparison')
    ax4.legend(loc='upper left')
    ax4b.legend(loc='lower right')

    # Adjust layout to remove excessive free space
    plt.tight_layout()

    # Save the figure using Path from pathlib
    output_path = Path("output") / "scalability_plots.png"
    output_path.parent.mkdir(parents=True, exist_ok=True)  # Create the directory if it doesn't exist
    plt.savefig(output_path, bbox_inches='tight')  # Save figure with tight bounding box
    plt.close()


def plot_train_complexity():
    # Example data: Replace these with the actual data from the callbacks
    batch_sizes = [8, 64]  # Maximum feasible batch sizes for MulT and LinMulT
    training_times = [callback_mult.train_epoch_times, callback_linmult.train_epoch_times]
    validation_times = [callback_mult.validation_epoch_times, callback_linmult.validation_epoch_times]
    peak_memory = [max(callback_mult.peak_memory_usage), max(callback_linmult.peak_memory_usage)]
    flops = [sum(callback_mult.flops_per_epoch), sum(callback_linmult.flops_per_epoch)]

    # Create subplots
    fig, axs = plt.subplots(2, 2, figsize=(14, 10))

    # Plot 1: Training Time per Epoch
    axs[0, 0].bar(['MulT', 'LinMulT'], [np.mean(training_times[0]), np.mean(training_times[1])], color=['blue', 'green'])
    axs[0, 0].set_title('Average Training Time per Epoch')
    axs[0, 0].set_ylabel('Time (seconds)')

    # Plot 2: Total Training Time
    axs[0, 1].bar(['MulT', 'LinMulT'], [sum(training_times[0]) + sum(validation_times[0]), sum(training_times[1]) + sum(validation_times[1])], color=['blue', 'green'])
    axs[0, 1].set_title('Total Training Time')
    axs[0, 1].set_ylabel('Time (seconds)')

    # Plot 3: Peak Memory Usage
    axs[1, 0].bar(['MulT', 'LinMulT'], peak_memory, color=['blue', 'green'])
    axs[1, 0].set_title('Peak Memory Usage')
    axs[1, 0].set_ylabel('Memory (MB)')

    # Plot 4: FLOPs Required for Training
    axs[1, 1].bar(['MulT', 'LinMulT'], flops, color=['blue', 'green'])
    axs[1, 1].set_title('FLOPs Required for Training')
    axs[1, 1].set_ylabel('FLOPs')

    # Adjust layout and show plot
    plt.tight_layout()
    plt.savefig('train_complexity.png')
    plt.close()


def plot_inference_batch():
    print('\nCreating plots about inference time across batch sizes...')

    with open('data/db_processed/fi/computational_complexity/inference_metrics_gpu_oowfr.pkl', 'rb') as f:
        gpu_data = pickle.load(f)

    with open('data/db_processed/fi/computational_complexity/inference_metrics_cpu_oowfr.pkl', 'rb') as f:
        cpu_data = pickle.load(f)

    batch_sizes = [1, 2, 4, 8]
    time_multiplier = 1.0

    # Prepare the data for plotting
    gpu_mult_mean = [item['inference_time_mean'] for item in gpu_data['MulT'][time_multiplier]]
    gpu_mult_std = [item['inference_time_std'] for item in gpu_data['MulT'][time_multiplier]]
    gpu_linmult_mean = [item['inference_time_mean'] for item in gpu_data['LinMulT'][time_multiplier]]
    gpu_linmult_std = [item['inference_time_std'] for item in gpu_data['LinMulT'][time_multiplier]]

    print('GPU MulT inference time mean for batch size 1, 2, 4, 8:', gpu_mult_mean)
    print('GPU MulT inference time std for batch size 1, 2, 4, 8:', gpu_mult_std)
    print('GPU LinMulT inference time std for batch size 1, 2, 4, 8:', gpu_linmult_mean)
    print('GPU LinMulT inference time std for batch size 1, 2, 4, 8:', gpu_linmult_std)

    cpu_mult_mean = [item['inference_time_mean'] for item in cpu_data['MulT'][time_multiplier]]
    cpu_mult_std = [item['inference_time_std'] for item in cpu_data['MulT'][time_multiplier]]
    cpu_linmult_mean = [item['inference_time_mean'] for item in cpu_data['LinMulT'][time_multiplier]]
    cpu_linmult_std = [item['inference_time_std'] for item in cpu_data['LinMulT'][time_multiplier]]

    print('CPU MulT inference time mean for batch size 1, 2, 4, 8:', cpu_mult_mean)
    print('CPU MulT inference time std for batch size 1, 2, 4, 8:', cpu_mult_std)
    print('CPU LinMulT inference time std for batch size 1, 2, 4, 8:', cpu_linmult_mean)
    print('CPU LinMulT inference time std for batch size 1, 2, 4, 8:', cpu_linmult_std)

    gpu_speedup = [speedup_ratio(gpu_mult_mean[i], gpu_linmult_mean[i]) for i in range(len(gpu_mult_mean))]
    cpu_speedup = [speedup_ratio(cpu_mult_mean[i], cpu_linmult_mean[i]) for i in range(len(cpu_mult_mean))]

    print('GPU speedup of LinMulT compared to MulT for batch size 1, 2, 4, 8:',gpu_speedup)
    print('CPU speedup of LinMulT compared to MulT for batch size 1, 2, 4, 8:',cpu_speedup)

    # Create the figure and subplots
    fig, axs = plt.subplots(1, 2, figsize=(14, 5))

    # Left subplot: Inference time as a bar plot
    axis_label_size = 14
    title_size = 16
    bar_width = 0.30
    space_between_bars = 0.03
    indices = np.arange(len(batch_sizes))

    axs[0].bar(indices[:len(gpu_mult_mean)], gpu_mult_mean, bar_width, yerr=gpu_mult_std, label='MulT', color=custom_blue, alpha=1.0, capsize=5, error_kw={'elinewidth': 2, 'capthick': 2}, edgecolor=border_blue, linewidth=line_width)
    axs[0].bar(indices[:len(gpu_linmult_mean)] + bar_width + space_between_bars, gpu_linmult_mean, bar_width, yerr=gpu_linmult_std, label='LinMulT', color=custom_green, alpha=1.0, capsize=5, error_kw={'elinewidth': 2, 'capthick': 2}, edgecolor=border_green, linewidth=line_width)

    for i in range(len(gpu_mult_mean)):
        speedup_val = f"{gpu_speedup[i]:.2f}x"
        height = max(gpu_mult_mean[i], gpu_linmult_mean[i])
        axs[0].text(indices[i] + (bar_width + space_between_bars) / 2, height + 15, speedup_val, ha='center', fontsize=12)

    axs[0].set_xlabel('Batch size', fontsize=axis_label_size)
    axs[0].set_ylabel('Inference time (ms)', fontsize=axis_label_size)
    axs[0].set_ylim([0,400])
    axs[0].set_xticks(indices[:len(gpu_linmult_mean)] + (bar_width + space_between_bars)/2)
    axs[0].set_xticklabels(batch_sizes[:len(gpu_linmult_mean)])
    axs[0].set_title('GPU inference time with speedup across batch sizes', fontsize=title_size)
    axs[0].legend(loc='upper left', framealpha=1.0)
    axs[0].grid(True)

    # Plotting bar for CPU inference time
    axs[1].bar(indices[:len(cpu_mult_mean)], cpu_mult_mean, bar_width, yerr=cpu_mult_std, label='MulT', color=custom_blue, alpha=1.0, capsize=5, error_kw={'elinewidth': 2, 'capthick': 2}, edgecolor=border_blue, linewidth=line_width)
    axs[1].bar(indices[:len(cpu_linmult_mean)] + bar_width + space_between_bars, cpu_linmult_mean, bar_width, yerr=cpu_linmult_std, label='LinMulT', color=custom_green, alpha=1.0, capsize=5, error_kw={'elinewidth': 2, 'capthick': 2}, edgecolor=border_green, linewidth=line_width)

    for i in range(len(cpu_mult_mean)):
        speedup_val = f"{cpu_speedup[i]:.2f}x"
        height = max(cpu_mult_mean[i], cpu_linmult_mean[i])
        axs[1].text(indices[i] + (bar_width + space_between_bars) / 2, height + 150, speedup_val, ha='center', fontsize=12)

    axs[1].set_xlabel('Batch size', fontsize=axis_label_size)
    axs[1].set_ylabel('Inference time (ms)', fontsize=axis_label_size)
    axs[1].set_yticks(range(0, 7500, 500))
    axs[1].set_xticks(indices[:len(cpu_linmult_mean)] + (bar_width + space_between_bars)/2)
    axs[1].set_xticklabels(batch_sizes[:len(cpu_linmult_mean)])
    axs[1].set_title('CPU inference time with speedup across batch sizes', fontsize=title_size)
    axs[1].legend(loc='upper left', framealpha=1.0)
    axs[1].grid(True)

    # Adjust layout and show the plot
    plt.tight_layout(pad=2.0)
    plt.savefig("output/plot_inference_batch.png")
    plt.close()


def plot_inference_window():
    print('\nCreating plots about inference time across window sizes...')

    with open('data/db_processed/fi/computational_complexity/inference_metrics_gpu_oowfr.pkl', 'rb') as f:
        gpu_data = pickle.load(f)

    with open('data/db_processed/fi/computational_complexity/inference_metrics_cpu_oowfr.pkl', 'rb') as f:
        cpu_data = pickle.load(f)

    window_sizes = [7.5, 15, 22.5, 30]
    ws = [0.5, 1.0, 1.5, 2.0]

    # Prepare the data for plotting
    gpu_mult_mean = [gpu_data['MulT'][w][0]['inference_time_mean'] for w in ws]
    gpu_mult_std = [gpu_data['MulT'][w][0]['inference_time_std'] for w in ws]
    gpu_linmult_mean = [gpu_data['LinMulT'][w][0]['inference_time_mean'] for w in ws]
    gpu_linmult_std = [gpu_data['LinMulT'][w][0]['inference_time_std'] for w in ws]

    print('GPU MulT inference time mean for window size 7.5s, 15s, 22.5s, 30s:', gpu_mult_mean)
    print('GPU MulT inference time std for window size 7.5s, 15s, 22.5s, 30s:', gpu_mult_std)
    print('GPU LinMulT inference time std for window size 7.5s, 15s, 22.5s, 30s:', gpu_linmult_mean)
    print('GPU LinMulT inference time std for window size 7.5s, 15s, 22.5s, 30s:', gpu_linmult_std)

    cpu_mult_mean = [cpu_data['MulT'][w][0]['inference_time_mean'] for w in ws]
    cpu_mult_std = [cpu_data['MulT'][w][0]['inference_time_std'] for w in ws]
    cpu_linmult_mean = [cpu_data['LinMulT'][w][0]['inference_time_mean'] for w in ws]
    cpu_linmult_std = [cpu_data['LinMulT'][w][0]['inference_time_std'] for w in ws]

    print('CPU MulT inference time mean for window size 7.5s, 15s, 22.5s, 30s:', cpu_mult_mean)
    print('CPU MulT inference time std for window size 7.5s, 15s, 22.5s, 30s:', cpu_mult_std)
    print('CPU LinMulT inference time std for window size 7.5s, 15s, 22.5s, 30s:', cpu_linmult_mean)
    print('CPU LinMulT inference time std for window size 7.5s, 15s, 22.5s, 30s:', cpu_linmult_std)

    gpu_speedup = [speedup_ratio(gpu_mult_mean[i], gpu_linmult_mean[i]) for i in range(len(gpu_mult_mean))]
    cpu_speedup = [speedup_ratio(cpu_mult_mean[i], cpu_linmult_mean[i]) for i in range(len(cpu_mult_mean))]

    print('GPU speedup of LinMulT compared to MulT for window size 7.5s, 15s, 22.5s, 30s:', gpu_speedup)
    print('CPU speedup of LinMulT compared to MulT for window size 7.5s, 15s, 22.5s, 30s:', cpu_speedup)

    # Create the figure and subplots
    fig, axs = plt.subplots(1, 2, figsize=(14, 5))

    # Left subplot: Inference time as a bar plot
    axis_label_size = 14
    title_size = 16
    bar_width = 0.30
    space_between_bars = 0.03
    indices = np.arange(len(window_sizes))

    axs[0].bar(indices[:len(gpu_mult_mean)], gpu_mult_mean, bar_width, yerr=gpu_mult_std, label='MulT', color=custom_blue, alpha=1.0, capsize=5, error_kw={'elinewidth': 2, 'capthick': 2}, edgecolor=border_blue, linewidth=line_width)
    axs[0].bar(indices[:len(gpu_linmult_mean)] + bar_width + space_between_bars, gpu_linmult_mean, bar_width, yerr=gpu_linmult_std, label='LinMulT', color=custom_green, alpha=1.0, capsize=5, error_kw={'elinewidth': 2, 'capthick': 2}, edgecolor=border_green, linewidth=line_width)

    for i in range(len(gpu_mult_mean)):
        speedup_val = f"{gpu_speedup[i]:.2f}x"
        height = max(gpu_mult_mean[i], gpu_linmult_mean[i])
        axs[0].text(indices[i] + (bar_width + space_between_bars) / 2, height + 12, speedup_val, ha='center', fontsize=12)

    axs[0].set_xlabel('Window size (s)', fontsize=axis_label_size)
    axs[0].set_ylabel('Inference time (ms)', fontsize=axis_label_size)
    axs[0].set_ylim([0,200])
    axs[0].set_xticks(indices[:len(gpu_linmult_mean)] + (bar_width + space_between_bars)/2)
    axs[0].set_xticklabels(window_sizes[:len(gpu_linmult_mean)])
    axs[0].set_title('GPU inference time with speedup across batch sizes', fontsize=title_size)
    axs[0].legend(loc='upper left', framealpha=1.0)
    axs[0].grid(True)

    # Plotting bar for CPU inference time
    axs[1].bar(indices[:len(cpu_mult_mean)], cpu_mult_mean, bar_width, yerr=cpu_mult_std, label='MulT', color=custom_blue, alpha=1.0, capsize=5, error_kw={'elinewidth': 2, 'capthick': 2}, edgecolor=border_blue, linewidth=line_width)
    axs[1].bar(indices[:len(cpu_linmult_mean)] + bar_width + space_between_bars, cpu_linmult_mean, bar_width, yerr=cpu_linmult_std, label='LinMulT', color=custom_green, alpha=1.0, capsize=5, error_kw={'elinewidth': 2, 'capthick': 2}, edgecolor=border_green, linewidth=line_width)

    for i in range(len(cpu_mult_mean)):
        speedup_val = f"{cpu_speedup[i]:.2f}x"
        height = max(cpu_mult_mean[i], cpu_linmult_mean[i])
        axs[1].text(indices[i] + (bar_width + space_between_bars) / 2, height + 250, speedup_val, ha='center', fontsize=12)

    axs[1].set_xlabel('Window size (s)', fontsize=axis_label_size)
    axs[1].set_ylabel('Inference time (ms)', fontsize=axis_label_size)
    axs[1].set_yticks(range(0, 7000, 500))
    axs[1].set_xticks(indices[:len(cpu_linmult_mean)] + (bar_width + space_between_bars)/2)
    axs[1].set_xticklabels(window_sizes[:len(cpu_linmult_mean)])
    axs[1].set_title('CPU inference time with speedup across window sizes', fontsize=title_size)
    axs[1].legend(loc='upper left', framealpha=1.0)
    axs[1].grid(True)
    
    # Adjust layout and show the plot
    plt.tight_layout(pad=2.0)
    plt.savefig("output/plot_inference_window.png")
    plt.close()


def plot_memory_comparison():

    with open('data/db_processed/fi/computational_complexity/inference_metrics_gpu_oowfr.pkl', 'rb') as f:
        gpu_data = pickle.load(f)

    # Sequence sizes (time multipliers)
    batch_sizes = [1, 2, 4, 8]
    sequence_multipliers = [0.5, 1.0, 1.5, 2.0]
    window_sizes = [7.5, 15, 22.5, 30]

    print('batch sizes:', batch_sizes)
    print('window sizes:', window_sizes)

    # Prepare data for allocated and reserved memory - batch size
    batch_mult_allocated = [gpu_data['MulT'][1.0][i]['peak_memory_allocated'] for i in range(len(batch_sizes))]
    batch_linmult_allocated = [gpu_data['LinMulT'][1.0][i]['peak_memory_allocated'] for i in range(len(batch_sizes))]

    batch_mult_reserved = [gpu_data['MulT'][1.0][i]['peak_memory_reserved'] for i in range(len(batch_sizes))]
    batch_linmult_reserved = [gpu_data['LinMulT'][1.0][i]['peak_memory_reserved'] for i in range(len(batch_sizes))]

    print('left plot: window size is 15s, batch size is changing')
    print('mult peak memory allocated:', batch_mult_allocated)
    print('linmult peak memory allocated:', batch_linmult_allocated)
    print('mult peak memory reserved:', batch_mult_reserved)
    print('linmult peak memory reserved:', batch_linmult_reserved)

    # data for sequence length
    mult_allocated = [gpu_data['MulT'][tm][0]['peak_memory_allocated'] for tm in sequence_multipliers]
    linmult_allocated = [gpu_data['LinMulT'][tm][0]['peak_memory_allocated'] for tm in sequence_multipliers]

    mult_reserved = [gpu_data['MulT'][tm][0]['peak_memory_reserved'] for tm in sequence_multipliers]
    linmult_reserved = [gpu_data['LinMulT'][tm][0]['peak_memory_reserved'] for tm in sequence_multipliers]

    print('right plot: window size is changing, batch size is 1')
    print('mult peak memory allocated:', mult_allocated)
    print('linmult peak memory allocated:', linmult_allocated)
    print('mult peak memory reserved:', mult_reserved)
    print('linmult peak memory reserved:', linmult_reserved)

    # Calculate percentage reduction
    batch_allocated_reduction = [(1 - batch_linmult_allocated[i] / batch_mult_allocated[i]) * 100 for i in range(len(batch_sizes))]
    batch_reserved_reduction = [(1 - batch_linmult_reserved[i] / batch_mult_reserved[i]) * 100 for i in range(len(batch_sizes))]

    print('percentage reduction for left plot')
    print('allocated reduction:', batch_allocated_reduction)
    print('reserved reduction:', batch_reserved_reduction)

    allocated_reduction = [(1 - linmult_allocated[i] / mult_allocated[i]) * 100 for i in range(len(sequence_multipliers))]
    reserved_reduction = [(1 - linmult_reserved[i] / mult_reserved[i]) * 100 for i in range(len(sequence_multipliers))]

    print('percentage reduction for right plot')
    print('allocated reduction:', allocated_reduction)
    print('reserved reduction:', reserved_reduction)

    # Bar plot width and indices
    axis_label_size = 14
    title_size = 16
    bar_width = 0.30
    space_between_bars = 0.03

    fig, axs = plt.subplots(1, 2, figsize=(14, 5))


    indices = np.arange(len(batch_sizes))

    # Overlay reserved memory with the same alpha
    axs[0].bar(indices, batch_mult_reserved, bar_width, label='MulT (Reserved)', color=custom_blue, alpha=1.0, hatch='//', edgecolor=border_blue, linewidth=line_width)
    axs[0].bar(indices + bar_width + space_between_bars, batch_linmult_reserved, bar_width, label='LinMulT (Reserved)', color=custom_green, alpha=1.0, hatch='//', edgecolor=border_green, linewidth=line_width)

    # Plot allocated memory
    axs[0].bar(indices, batch_mult_allocated, bar_width, label='MulT (Allocated)', color=custom_blue, alpha=1.0, edgecolor=border_blue, linewidth=line_width)
    axs[0].bar(indices + bar_width + space_between_bars, batch_linmult_allocated, bar_width, label='LinMulT (Allocated)', color=custom_green, alpha=1.0, edgecolor=border_green, linewidth=line_width)

    # Add percentage reduction text on top of bars
    for i in range(len(batch_mult_allocated)):
        height = max(batch_mult_reserved[i], batch_linmult_reserved[i])
        reduction_text = f"{batch_allocated_reduction[i]:.2f}% [{batch_reserved_reduction[i]:.2f}%]"
        axs[0].text(indices[i] + (bar_width + space_between_bars) / 2, height + 20, reduction_text, ha='center', fontsize=11)

    axs[0].set_xlabel('Batch size', fontsize=axis_label_size)
    axs[0].set_ylabel('GPU memory usage (MB)', fontsize=axis_label_size)
    axs[0].set_ylim([0, 1700])
    axs[0].set_yticks(range(0, 1700, 100))
    axs[0].set_xticks(indices + (bar_width + space_between_bars) / 2)
    axs[0].set_xticklabels(batch_sizes)
    axs[0].set_title('GPU memory usage across batch sizes', fontsize=title_size)
    axs[0].legend(loc='upper left', framealpha=1.0)
    axs[0].grid(True)


    indices = np.arange(len(sequence_multipliers))

    # Overlay reserved memory with the same alpha
    axs[1].bar(indices, mult_reserved, bar_width, label='MulT (Reserved)', color=custom_blue, alpha=1.0, hatch='//', edgecolor=border_blue, linewidth=line_width)
    axs[1].bar(indices + bar_width + space_between_bars, linmult_reserved, bar_width, label='LinMulT (Reserved)', color=custom_green, alpha=1.0, hatch='//', edgecolor=border_green, linewidth=line_width)

    # Plot allocated memory
    axs[1].bar(indices, mult_allocated, bar_width, label='MulT (Allocated)', color=custom_blue, alpha=1.0, edgecolor=border_blue, linewidth=line_width)
    axs[1].bar(indices + bar_width + space_between_bars, linmult_allocated, bar_width, label='LinMulT (Allocated)', color=custom_green, alpha=1.0, edgecolor=border_green, linewidth=line_width)

    # Add percentage reduction text on top of bars
    for i in range(len(mult_allocated)):
        height = max(mult_reserved[i], linmult_reserved[i])
        reduction_text = f"{allocated_reduction[i]:.2f}% [{reserved_reduction[i]:.2f}%]"
        axs[1].text(indices[i] + (bar_width + space_between_bars) / 2, height + 20, reduction_text, ha='center', fontsize=11)

    axs[1].set_xlabel('Window size (s)', fontsize=axis_label_size)
    axs[1].set_ylabel('GPU memory usage (MB)', fontsize=axis_label_size)
    axs[1].set_ylim([0, 900])
    axs[1].set_yticks(range(0, 900, 100))
    axs[1].set_xticks(indices + (bar_width + space_between_bars) / 2)
    axs[1].set_xticklabels(window_sizes)
    axs[1].set_title('GPU memory usage across window sizes', fontsize=title_size)
    axs[1].legend(loc='upper left', framealpha=1.0)
    axs[1].grid(True)

    # Adjust layout and show the plot
    plt.tight_layout(pad=2.0)
    plt.savefig("output/plot_inference_memory.png")
    plt.close()


def plot_flops_window():
    with open('data/db_processed/fi/computational_complexity/training_flops_gpu_oowfr.pkl', 'rb') as f:
        flops_data = pickle.load(f)

    # Sequence sizes (time multipliers) and batch sizes
    sequence_multipliers = [0.5, 1.0, 1.5, 2.0]
    window_sizes = [7.5, 15, 22.5, 30]

    print('window sizes:', window_sizes)

    # Prepare data for inference and training FLOPs
    mult_inference_flops = [flops_data[tm]['MulT']['inference_flops_mean'] for tm in sequence_multipliers]
    linmult_inference_flops = [flops_data[tm]['LinMulT']['inference_flops_mean'] for tm in sequence_multipliers]

    mult_training_flops = [flops_data[tm]['MulT']['training_flops_mean'] for tm in sequence_multipliers]
    linmult_training_flops = [flops_data[tm]['LinMulT']['training_flops_mean'] for tm in sequence_multipliers]

    print('FLOPs calculation for a single sample during inference across window durations')
    print('mult flops:', mult_inference_flops)
    print('linmult flops:', linmult_inference_flops)
    print('FLOPs calculation for a single sample during training across window durations')
    print('mult flops:', mult_training_flops)
    print('linmult flops:', linmult_training_flops)

    # Calculate percentage reduction in FLOPs between MulT and LinMulT
    inference_flops_reduction = [(1 - linmult_inference_flops[i] / mult_inference_flops[i]) * 100 for i in range(len(sequence_multipliers))]
    training_flops_reduction = [(1 - linmult_training_flops[i] / mult_training_flops[i]) * 100 for i in range(len(sequence_multipliers))]

    print('percentage reduction in FLOPs:')
    print('inference:', inference_flops_reduction)
    print('training:', training_flops_reduction)

    # Plot settings
    bar_width = 0.30
    space_between_bars = 0.03
    axis_label_size = 14
    title_size = 16
    line_width = 1.5


    fig, axs = plt.subplots(1, 1, figsize=(7, 5))


    indices = np.arange(len(sequence_multipliers))

    # Overlay reserved memory with the same alpha
    axs.bar(indices, mult_training_flops, bar_width, label='MulT (training)', color=custom_blue, alpha=1.0, hatch='//', edgecolor=border_blue, linewidth=line_width)
    axs.bar(indices + bar_width + space_between_bars, linmult_training_flops, bar_width, label='LinMulT (training)', color=custom_green, alpha=1.0, hatch='//', edgecolor=border_green, linewidth=line_width)

    # Plot allocated memory
    axs.bar(indices, mult_inference_flops, bar_width, label='MulT (inference)', color=custom_blue, alpha=1.0, edgecolor=border_blue, linewidth=line_width)
    axs.bar(indices + bar_width + space_between_bars, linmult_inference_flops, bar_width, label='LinMulT (inference)', color=custom_green, alpha=1.0, edgecolor=border_green, linewidth=line_width)

    # Add percentage reduction text on top of bars
    for i in range(len(mult_inference_flops)):
        height = max(mult_training_flops[i], linmult_training_flops[i])
        reduction_text = f"{inference_flops_reduction[i]:.2f}% [{training_flops_reduction[i]:.2f}%]"
        axs.text(indices[i] + (bar_width + space_between_bars) / 2, height + 0.05e+11, reduction_text, ha='center', fontsize=11)

    axs.set_xlabel('Window size (s)', fontsize=axis_label_size)
    axs.set_ylabel('FLOPs', fontsize=axis_label_size)
    axs.set_ylim([0, 2.5e+11])
    axs.set_yticks(np.arange(0, 2.5e+11 + 1, 0.25e+11))
    axs.set_xticks(indices + (bar_width + space_between_bars) / 2)
    axs.set_xticklabels(window_sizes)
    axs.set_title('FLOPs across window sizes', fontsize=title_size)
    axs.legend(loc='upper left', framealpha=1.0)
    axs.grid(True)

    # Adjust layout and show the plot
    plt.tight_layout(pad=2.0)
    plt.savefig("output/plot_flops.png")
    plt.close()


if __name__ == "__main__":
    plot_flops_window()
    #plot_inference_batch()
    #plot_inference_window()
    #plot_memory_comparison()



from pathlib import Path
from tqdm import tqdm
import numpy as np
import pandas as pd
import pickle
from exordium.text.roberta import RobertaWrapper


DB = Path("data/db_processed/mosei")
DB_TEXT = DB / "roberta"
LABEL_PATH = DB / 'mosei_label.csv'


if __name__ == "__main__":
    df = pd.read_csv(LABEL_PATH)

    roberta = RobertaWrapper()

    d = {}
    for index, row in tqdm(df.iterrows(), total=len(df)):
        text = row['text'].lower()
        feature = roberta(text)
        feature = feature.squeeze(0)
        video_id = row['video_id']
        clip_id = row['clip_id']
        
        if video_id not in d:
            d[video_id] = {}

        d[video_id][clip_id] = feature.numpy()
    
    DB_TEXT.mkdir(parents=True, exist_ok=True)
    with open(DB_TEXT / 'mosei_roberta.pkl', 'wb') as f:
        pickle.dump(d, f)


import time
import numpy as np
import torch
import torch.profiler


def measure_inference_time(model, batch):
    # Warm-up to stabilize performance
    for _ in range(10):
        with torch.no_grad():
            _ = model(batch)

    latency_list = []
    for _ in range(10):
        # Measure inference time
        torch.cuda.synchronize()
        start_time = time.time()

        with torch.no_grad():
            _ = model(batch)

        torch.cuda.synchronize()
        latency = (time.time() - start_time) * 1000  # Convert to ms
        latency_list.append(latency)

    latency_array = np.array(latency_list)
    return np.mean(latency_array), np.std(latency_array)


def percentage_reduction(original, new):
    if original == 0: return 0
    return ((original - new) / original) * 100


def speedup_ratio(time1, time2):
    if time2 == 0: return 0
    return time1 / time2


def measure_memory_usage(model, batch, device: str):
    torch.cuda.reset_peak_memory_stats(device)

    with torch.no_grad():
        _ = model(batch)

    peak_memory_allocated = torch.cuda.max_memory_allocated(device) / (1024 * 1024)  # Convert to MB
    peak_memory_reserved = torch.cuda.max_memory_reserved(device) / (1024 * 1024)  # Convert to MB

    return peak_memory_allocated, peak_memory_reserved


def measure_flops_with_profiler(model, batch):
    # Use torch.profiler to measure FLOPs
    with torch.profiler.profile(
        activities=[
            torch.profiler.ProfilerActivity.CPU,
            torch.profiler.ProfilerActivity.CUDA
        ],
        record_shapes=True,
        with_flops=True
    ) as profiler:
        with torch.no_grad():
            model(batch)

    # Print the profiler key metrics
    flops = profiler.key_averages().total_average().flops
    return flops


def _measure_training_flops_with_profiler(model, batch, target, loss_fn):
    # Enable profiling to measure both forward and backward FLOPs
    with torch.profiler.profile(
        activities=[
            torch.profiler.ProfilerActivity.CPU,
            torch.profiler.ProfilerActivity.CUDA
        ],
        record_shapes=True,
        with_flops=True
    ) as profiler:
        
        # Run the forward pass
        output = model(batch)
        
        # Calculate the loss
        loss = loss_fn(output, target)

        # Backward pass to calculate gradients
        loss.backward()

    # Get the total FLOPs for both forward and backward passes (training FLOPs)
    flops = profiler.key_averages().total_average().flops
    return flops


import time
import pickle
from pathlib import Path
from tqdm import tqdm
import numpy as np
from datetime import datetime
import torch
from torch.profiler import profile, ProfilerActivity
import lightning as L
from lightning.pytorch.callbacks import ModelCheckpoint
from lightning.pytorch.loggers import CSVLogger
from linmult import LinMulT
from performance_metrics import measure_memory_usage
from personalitylinmult.train.train import ModelWrapper, load_config
from personalitylinmult.train.fi import OOWFRDataModule
from utils import load_mult, load_linmult, load_sample
from performance_metrics import speedup_ratio
from exordium.utils.loss import bell_l2_l1_loss


def measure_inference_flops(model, batch):
    with profile(
        activities=[ProfilerActivity.CPU, ProfilerActivity.CUDA],
        record_shapes=True,
        with_flops=True
    ) as prof:
        with torch.no_grad():
            model(batch)
    flops = prof.key_averages().total_average().flops
    return flops

def measure_training_flops(model, batch, target):
    optimizer = torch.optim.Adam(model.parameters(), lr=0.001)
    loss_fn = bell_l2_l1_loss
    model.train()
    with profile(
        activities=[ProfilerActivity.CPU, ProfilerActivity.CUDA],
        record_shapes=True,
        with_flops=True
    ) as prof:
        optimizer.zero_grad()
        output = model(batch)[0] # mult and linmult specific
        loss = loss_fn(output, target)
        loss.backward()
        optimizer.step()
    flops = prof.key_averages().total_average().flops
    return flops

def measure_inference_flops_and_flops(model, batch):
    # Measure FLOPs
    inference_flops = measure_inference_flops(model, batch)
    
    # Measure time
    torch.cuda.synchronize() if torch.cuda.is_available() else None
    start_time = time.time()
    
    with torch.no_grad():
        model(batch)
    
    torch.cuda.synchronize() if torch.cuda.is_available() else None
    end_time = time.time()
    
    inference_time = end_time - start_time  # In seconds
    flops_per_sec = inference_flops / inference_time if inference_time > 0 else float('inf')
    
    return inference_flops, inference_time, flops_per_sec

def measure_training_flops_and_flops(model, batch, target):
    optimizer = torch.optim.Adam(model.parameters(), lr=0.001)
    loss_fn = bell_l2_l1_loss
    model.train()
    
    # Measure FLOPs
    training_flops = measure_training_flops(model, batch, target)
    
    # Measure time
    torch.cuda.synchronize() if torch.cuda.is_available() else None
    start_time = time.time()
    
    optimizer.zero_grad()
    output = model(batch)[0] # mult and linmult specific
    loss = loss_fn(output, target)
    loss.backward()
    optimizer.step()
    
    torch.cuda.synchronize() if torch.cuda.is_available() else None
    end_time = time.time()
    
    training_time = end_time - start_time  # In seconds
    flops_per_sec = training_flops / training_time if training_time > 0 else float('inf')
    
    return training_flops, training_time, flops_per_sec

def average_measurements(func, num_runs=10):
    flops_list = []
    flops_per_sec_list = []
    
    for _ in range(10):
        func()
    
    for _ in tqdm(range(num_runs)):
        flops, exec_time, flops_per_sec = func()
        flops_list.append(flops)
        flops_per_sec_list.append(flops_per_sec)
    
    flops_mean = np.mean(flops_list)
    flops_per_sec_mean = np.mean(flops_per_sec_list)
    return flops_mean, flops_per_sec_mean

def train_flops_experiment():
    torch.set_float32_matmul_precision('highest') 
    device = torch.device('cuda:0')
    mult = load_mult().to(device)
    linmult = load_linmult().to(device)
    target = torch.rand((1,5), device=device)

    # Measure Inference FLOPs and FLOPS
    def inference_mult_func():
        return measure_inference_flops_and_flops(mult, batch)

    def inference_linmult_func():
        return measure_inference_flops_and_flops(linmult, batch)

    # Measure Training FLOPs and FLOPS
    def training_mult_func():
        return measure_training_flops_and_flops(mult, batch, target)

    def training_linmult_func():
        return measure_training_flops_and_flops(linmult, batch, target)

    output = {}
    for tm in [0.5, 1.0, 1.5, 2.0]:
        batch = load_sample(features='oowfr', batch_size=1, time_multiplier=tm)
        batch = [torch.FloatTensor(elem).to(device) for elem in batch]

        mult_inference_flops_mean, \
            mult_inference_flops_per_sec_mean = average_measurements(
                inference_mult_func, num_runs=10)
        print(f"MulT inference FLOPs mean: {mult_inference_flops_mean:.2e}")
        print(f"MulT inference FLOPS mean: {mult_inference_flops_per_sec_mean:.2e}")
        linmult_inference_flops_mean, \
            linmult_inference_flops_per_sec_mean = average_measurements(
                inference_linmult_func, num_runs=10)
        print(f"LinMulT inference FLOPs mean: {linmult_inference_flops_mean:.2e}")
        print(f"LinMulT inference FLOPS mean: {linmult_inference_flops_per_sec_mean:.2e}")

        speedup_inference_flops = speedup_ratio(mult_inference_flops_mean, linmult_inference_flops_mean)
        speedup_inference_flops_per_sec = speedup_ratio(mult_inference_flops_per_sec_mean, linmult_inference_flops_per_sec_mean)
        print('speedup inference FLOPs:', speedup_inference_flops)
        print('speedup inference FLOPS:', speedup_inference_flops_per_sec)

        mult_training_flops_mean, \
            mult_training_flops_per_sec_mean = average_measurements(
                training_mult_func, num_runs=10)
        print(f"MulT training FLOPs mean: {mult_training_flops_mean:.2e}")
        print(f"MulT training FLOPS mean: {mult_training_flops_per_sec_mean:.2e}")
        linmult_training_flops_mean, \
            linmult_training_flops_per_sec_mean = average_measurements(
                training_linmult_func, num_runs=10)
        print(f"LinMulT training FLOPs mean: {linmult_training_flops_mean:.2e}")
        print(f"LinMulT training FLOPS mean: {linmult_training_flops_per_sec_mean:.2e}")

        output[tm] = {}
        output[tm]['MulT'] = {
            'inference_flops_mean': mult_inference_flops_mean,
            'inference_flops_per_sec_mean': mult_inference_flops_per_sec_mean,
            'training_flops_mean': mult_training_flops_mean,
            'training_flops_per_sec_mean': mult_training_flops_per_sec_mean,
        }
        output[tm]['LinMulT'] = {
            'inference_flops_mean': linmult_inference_flops_mean,
            'inference_flops_per_sec_mean': linmult_inference_flops_per_sec_mean,
            'training_flops_mean': linmult_training_flops_mean,
            'training_flops_per_sec_mean': linmult_training_flops_per_sec_mean,
        }

    with open('data/db_processed/fi/computational_complexity/training_flops_gpu_oowfr.pkl', 'wb') as f:
        pickle.dump(output, f)


if __name__ == "__main__":
    train_flops_experiment()


import pickle
from tqdm import tqdm
import torch
from performance_metrics import (
    measure_inference_time, percentage_reduction, speedup_ratio,
    measure_memory_usage, measure_flops_with_profiler
)
from torch.utils.data import Dataset, DataLoader
from utils import load_mult, load_linmult, load_sample


def run_inference_experiment(model, batch, device: str):
    torch.cuda.empty_cache()
    model = model.to(device)
    model.eval()
    batch = [torch.FloatTensor(inp).to(device) for inp in batch]

    # Measure inference time
    inference_time_mean, inference_time_std = measure_inference_time(model, batch)

    # Measure memory usage
    peak_memory_allocated, peak_memory_reserved = measure_memory_usage(model, batch, device)

    # Measure FLOPS
    flops = measure_flops_with_profiler(model, batch)

    del batch

    return {
        'inference_time_mean': inference_time_mean,
        'inference_time_std': inference_time_std,
        'peak_memory_allocated': peak_memory_allocated,
        'peak_memory_reserved': peak_memory_reserved,
        'flops': flops
    }


def run_comparative_experiment(mult_model, linmult_model, batch, device1, device2):

    # Run experiments for LinMulT
    print(f'LinMulT run_inference_experiment on {device2}')
    linmult_results = run_inference_experiment(linmult_model, batch, device2)

    # Run experiments for MulT
    print(f'MulT run_inference_experiment on {device1}')
    mult_results = run_inference_experiment(mult_model, batch, device1)

    # Calculate metrics
    speedup = speedup_ratio(mult_results['inference_time_mean'], linmult_results['inference_time_mean'])
    time_reduction = percentage_reduction(mult_results['inference_time_mean'], linmult_results['inference_time_mean'])
    memory_allocated_reduction = percentage_reduction(mult_results['peak_memory_allocated'], linmult_results['peak_memory_allocated'])
    memory_reserved_reduction = percentage_reduction(mult_results['peak_memory_reserved'], linmult_results['peak_memory_reserved'])

    # Display results
    print(f"[MulT]    Inference Time: {mult_results['inference_time_mean']:.2f} \u00B1 {mult_results['inference_time_std']:.2f} ms")
    print(f"[LinMulT] Inference Time: {linmult_results['inference_time_mean']:.2f} \u00B1 {linmult_results['inference_time_std']:.2f} ms")
    print(f"[LinMulT] Speedup Ratio: {speedup:.2f}x")
    print(f"[LinMulT] Time Reduction: {time_reduction:.2f}%")

    print(f"\n[MulT]    Allocated Memory Usage: {mult_results['peak_memory_allocated']:.2f} MB")
    print(f"[LinMulT] Allocated Memory Usage: {linmult_results['peak_memory_allocated']:.2f} MB")
    print(f"[LinMulT] Allocated Memory Reduction: {memory_allocated_reduction:.2f}%")

    print(f"\n[MulT]    Reserved Memory Usage: {mult_results['peak_memory_reserved']:.2f} MB")
    print(f"[LinMulT] Reserved Memory Usage: {linmult_results['peak_memory_reserved']:.2f} MB")
    print(f"[LinMulT] Reserved Memory Reduction: {memory_reserved_reduction:.2f}%")

    print(f"\n[MulT]    FLOPs: {mult_results['flops']:.2e}")
    print(f"[LinMulT] FLOPs: {linmult_results['flops']:.2e}")

    return mult_results, linmult_results


def find_max_feasible_batch_size(model, device: str):
    torch.cuda.empty_cache()
    model = model.to(device)
    model.eval()

    for time_multiplier in [0.5, 1.0, 1.5]:
        print('Time_multiplier:', time_multiplier)
        for batch_size in [1, 2, 4, 8, 16, 32, 64, 128, 256, 512, 1024]:
            batch = load_sample(batch_size=batch_size, time_multiplier=time_multiplier)
            input_data = [torch.FloatTensor(inp).to(device) for inp in batch]

            try:
                with torch.no_grad():
                    output = model(input_data)
                print('Batch size:', input_data[0].shape[0])
            except:
                print('Batch size:', input_data[0].shape[0], 'CUDA OOM!')
            
            del input_data


if __name__ == "__main__":
    torch.set_float32_matmul_precision('highest') # this was the only option at the time of publication 
    device1 = torch.device('cuda:0')
    device2 = torch.device('cuda:1')
    device = device2

    mult = load_mult()
    linmult = load_linmult()

    #print('[MulT] Find maximum feasible batch size for inference...')
    #find_max_feasible_batch_size(mult, device1)
    #print('[LinMulT] Find maximum feasible batch size for inference...')
    #find_max_feasible_batch_size(linmult, device2)

    features = 'oowfr' # oowfr
    mult_results_dict = {}
    linmult_results_dict = {}
    
    for time_multiplier in [0.5, 1.0, 1.5, 2.0]:
        print(f'{"-"*20}\nTime multiplier ({time_multiplier})')
        mult_results_dict[time_multiplier] = []
        linmult_results_dict[time_multiplier] = []

        for batch_size in [1, 2, 4, 8]:
            print(f'{"-"*20}\nBatch ({batch_size})')

            batch = load_sample(features=features, batch_size=batch_size, time_multiplier=time_multiplier)

            try:
                mult_results, linmult_results = run_comparative_experiment(mult, linmult, batch, device, device)
                mult_results_dict[time_multiplier].append(mult_results)
                linmult_results_dict[time_multiplier].append(linmult_results)
                del batch

                with open(f'data/db_processed/fi/computational_complexity/inference_metrics_gpu_{features}.pkl', 'wb') as f:
                    pickle.dump({'MulT': mult_results_dict, 'LinMulT': linmult_results_dict}, f)
            except:
                print('CUDA OOM!')

        '''
        for batch_size in [16, 32, 64, 128, 256, 512, 1024]:
            print(f'{"-"*20}\nBatch ({batch_size})')
            batch = load_sample(features=features, batch_size=batch_size, time_multiplier=time_multiplier)

            try:
                linmult_results = run_inference_experiment(linmult, batch, device2)
                print(f"[LinMulT] Inference Time: {linmult_results['inference_time_mean']:.2f} \u00B1 {linmult_results['inference_time_std']:.2f} ms")
                print(f"[LinMulT] Memory Usage - Allocated: {linmult_results['peak_memory_allocated']:.2f} MB")
                print(f"[LinMulT] Memory Usage - Reserved: {linmult_results['peak_memory_reserved']:.2f} MB")
                print(f"[LinMulT] FLOPs: {linmult_results['flops']:.2e}")
                linmult_results_list.append(linmult_results)
                del batch

                with open(f'data/db_processed/fi/computational_complexity/inference_metrics_{str(time_multiplier)}_{features}.pkl', 'wb') as f:
                    pickle.dump({'MulT': mult_results_list, 'LinMulT': linmult_results_list}, f)
            except:
                print('CUDA OOM!')
        '''


from pathlib import Path
from tqdm import tqdm
import numpy as np
import pandas as pd
import pickle
from exordium.video.openface import read_openface_au


DB = Path("data/db_processed/mosi")
LABEL_PATH = DB / 'mosi_label.csv'


if __name__ == "__main__":

    df = pd.read_csv(LABEL_PATH)

    with open(DB / 'roberta' / 'mosi_roberta.pkl', 'rb') as f:
        roberta_dict = pickle.load(f)

    for subset in ['train', 'valid', 'test']:

        df_subset = df[df['mode'] == subset]

        samples = {}
        for index, row in tqdm(df_subset.iterrows(), total=len(df_subset), desc=f'{subset}'):
            if row['mode'] != subset: continue

            video_id = row['video_id']
            clip_id = row['clip_id']

            # collect calculated features
            try:
                egemaps_lld = np.load(DB / 'egemaps_lld' / video_id / f'{clip_id}.npy') # (T, F)
                wav2vec2 = np.load(DB / 'wav2vec2' / video_id / f'{clip_id}.npy') # (T, F)
                _, au, _ = read_openface_au(DB / 'openface' / video_id / str(clip_id) / f'{clip_id}.csv') # (T, F)
                with open(DB / 'fabnet' / video_id / f'{clip_id}.pkl', 'rb') as f:
                    _, fabnet = pickle.load(f) # (T, F)
                roberta = roberta_dict[video_id][clip_id].numpy() # (T, F)
            except:
                print(f'Invalid sample: {video_id}/{clip_id}')
                continue

            if video_id not in samples:
                samples[video_id] = {}

            sample = [egemaps_lld, au, wav2vec2, fabnet, roberta]
            samples[video_id][clip_id] = {
                'egemaps_lld': egemaps_lld,
                'au': au,
                'wav2vec2': wav2vec2,
                'fabnet': fabnet,
                'roberta': roberta,
                'sentiment': float(row['label'])
            } # oowfr

            #print(f'{video_id}/{clip_id}', float(row['label']), egemaps_lld.shape, wav2vec2.shape, au.shape, fabnet.shape, roberta.shape)
            assert all([elem.ndim == 2 and elem.shape[0] != 0 for elem in sample])

        output_path = DB / 'oowfr' / f'mosi_{subset}_oowfr.pkl'
        output_path.parent.mkdir(parents=True, exist_ok=True)
        with open(output_path, 'wb') as f:
            pickle.dump(samples, f)
        



import pickle
from pathlib import Path
import numpy as np
import torch
import lightning as L
from torch.utils.data import Dataset, DataLoader
from linmult import LinMulT
from personalitylinmult.train.fi import OOWFRDataset


def load_mult():
    model_config = 'config/experiment_computational_complexity/OOWFR_MulT.yaml'
    model = LinMulT(model_config)
    return model


def load_linmult():
    model_config = 'config/experiment_computational_complexity/OOWFR_LinMulT.yaml'
    model = LinMulT(model_config)
    return model


def load_sample(db_root: str = 'data/db_processed/fi', features: str = 'oowfr', batch_size: int = 64, time_multiplier: float = 1.0):
    tmp_path = Path(db_root) / 'tmp'
    tmp_path.mkdir(parents=True, exist_ok=True)

    batch_path = tmp_path / f'fi_{features}_batch.pkl'
    if batch_path.exists():
        with open(batch_path, 'rb') as f:
            batch = pickle.load(f)

    else:
        print('Load OOWFR data...')
        if features == 'oowfr':
            ds = OOWFRDataset('test')
            dl = DataLoader(ds, batch_size=1024, shuffle=False)
            batch, _ = next(iter(dl))
            batch = [inp.cpu().detach().numpy() for inp in batch]
        else:
            batch = [
                np.random.rand(1024,1500,25),
                np.random.rand(1024,450,35),
                np.random.rand(1024,80,768)
            ]

        with open(batch_path, 'wb') as f:
            pickle.dump(batch, f)

    modified_tensors = []
    for tensor in batch:
        orig_batch_size, orig_time_dim, feature_dim = tensor.shape
        sliced_tensor = tensor[:batch_size, :, :]
        new_time_dim = int(orig_time_dim * time_multiplier)

        if new_time_dim < orig_time_dim:
            sliced_tensor = sliced_tensor[:, :new_time_dim, :]
        else:
            pad_length = new_time_dim - orig_time_dim
            padding = np.zeros((batch_size, pad_length, feature_dim))
            sliced_tensor = np.concatenate((sliced_tensor, padding), axis=1)

        modified_tensors.append(sliced_tensor)

    return modified_tensors


if __name__ == "__main__":
    sample_short = load_sample(batch_size=4, time_multiplier=0.5)
    sample_long = load_sample(batch_size=8, time_multiplier=1.5)
    print('short sample shape:', [elem.shape for elem in sample_short])
    print('long sample shape:', [elem.shape for elem in sample_long])


import os
import argparse
from pathlib import Path
import time
from tqdm import tqdm
from exordium.audio.smile import OpensmileWrapper
from exordium.audio.wav2vec import Wav2vec2Wrapper
from exordium.audio.io import video2audio

DB = Path("data/db_processed/mosei")
DB_RAW_VIDEOS = DB / "Raw"
DB_VIDEOS = DB / "Converted"
DB_AUDIOS = DB / "Audio"


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Script to preprocess MOSI acoustic features.")
    parser.add_argument('--start',  type=int, default=0, help='participant id slice start')
    parser.add_argument('--end',    type=int, default=23000, help='participant id slice end')
    args = parser.parse_args()

    for converted_video_path in tqdm(sorted(list(DB_VIDEOS.glob("**/*.mp4")))):
        video2audio(
            converted_video_path,
            DB_AUDIOS / converted_video_path.parent.name / f'{converted_video_path.stem}.wav',
            sr=16000
        )

    wav2vec_extractor = Wav2vec2Wrapper()
    opensmile_extractor = OpensmileWrapper('egemaps', 'lld')
    
    audio_paths = sorted(list(DB_AUDIOS.glob("**/*.wav")))[args.start:args.end]
    print('Number of audios:', len(audio_paths))

    for v, audio_path in tqdm(enumerate(audio_paths), total=len(audio_paths), desc='Audios'):
        start_time = time.time()
        video_name = audio_path.parent.name
        video_id = audio_path.stem

        try:
            wav2vec_feature = wav2vec_extractor.audio_to_feature(audio_path, output_path=DB / 'wav2vec2' / video_name / f'{video_id}.npy')

            smile_feature = opensmile_extractor.audio_to_feature(audio_path, output_path=DB / 'egemaps_lld' / video_name / f'{video_id}.npy')
        except Exception as e:
            with open(DB / "mosei_skip_audio.txt", "a") as f:
                f.write(f'{video_name},{video_id},{e}\n')

        end_time = time.time()
        print(f"Elapsed time: {end_time - start_time:03f} seconds")